import java.util.HashMap;
import java.util.Arrays;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

// ConcTest [dl]

class BankReadWriteLock {

        private static class Account {
            private int balance;
            ReentrantLock lock = new ReentrantLock();

            Account(int balance) { 
                this.balance = balance;
            }

            int balance() { return balance; }

            boolean deposit(int value) {
                balance += value;
                return true;
            }

            boolean withdraw(int value) {
                if (value > balance)
                    return false;
                balance -= value;
                return true;
            }
        }

        private ReentrantReadWriteLock lockBanco;
        private HashMap<Integer, Account> accounts;
        private int nextId;

        public BankReadWriteLock(){
            this.lockBanco = new ReentrantReadWriteLock();
            this.accounts = new HashMap<Integer, Account>();
            this.nextId = 0;
        }

        // create account and return account id
        public int createAccount(int balance) {
            Account account = new Account(balance);
            int id;
            lockBanco.writeLock().lock();

            try{
                id = nextId;
                nextId += 1;
                accounts.put(id, account);
            } finally {
                lockBanco.writeLock().unlock();
            }
            return id;
        }


        // Fails if no such account
        public boolean deposit(int id, int value) {
            Account account;
            lockBanco.readLock().lock();
            try {
                account = accounts.get(id);
                if (account == null)
                    return false;
                account.lock.lock();
            }finally {
                lockBanco.readLock().unlock();
            }
            try {
                return account.deposit(value);
            } finally {
                account.lock.unlock();
            }
        }

        // Fails if no such account or insufficient balance
        public boolean withdraw(int id, int value) {
            Account account;
            lockBanco.readLock().lock();
            try{
                account = accounts.get(id);
                if( account==null) 
                    return false;
                account.lock.lock();
            } finally {
                lockBanco.readLock().unlock();
            }
            try{
                return account.withdraw(value);
            } finally {
                account.lock.unlock();
            }
        }

        // Fails if either account does not exist or insufficient balance
        public boolean transfer(int from, int to, int value) {
            Account cfrom, cto;
            lockBanco.readLock().lock();
            try {
                cfrom = accounts.get(from);
                cto = accounts.get(to);
                if (cfrom == null || cto == null)
                    return false;
                cfrom.lock.lock();
                cto.lock.lock();
            } finally {
                lockBanco.readLock().unlock();
            }
            try {
                return cfrom.withdraw(value) && cto.deposit(value);

            } finally {
                cfrom.lock.unlock();
                cto.lock.unlock();
            }
        }

        public int totalBalance(int[] ids) {
            int total = 0;
            Account[] acs =new Account[ids.length];
            lockBanco.readLock().lock();
            try {
                for (int i : ids) {
                    Account account = this.accounts.get(i);
                    if (account == null)
                        return 0;
                    acs[i]=account;
                }
                for(Account account : acs){
                    account.lock.lock();
                }
            }finally {
                lockBanco.readLock().unlock();
            }
            for(Account account : acs){
                total += account.balance;
                account.lock.unlock();
            }
            return total;
        }
    
    public static void main (String[] args){
        BankReadWriteLock bank = new BankReadWriteLock();
        
        bank.createAccount(15);
        bank.createAccount(10);
        
        Thread t1 = new Thread(() -> { bank.transfer(0,1,5); });
        Thread t2 = new Thread(() -> { bank.transfer(1,0,10); });
        t1.start();
        t2.start();
        
        try {
            t1.join();
            t2.join();
        } catch(InterruptedException e) {e.printStackTrace();}
    }
}

